/*
 *  ircd-hybrid: an advanced Internet Relay Chat Daemon(ircd).
 *  dh.c: Diffie-Hellman key exchange.
 *
 *  Copyright (C) 2003 by the past and present ircd coders, and others.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  $Id: dh.c,v 1.2 2004/02/06 23:45:50 ircd Exp $
 */


#include "stdinc.h"
#include "setup.h"
#include "config.h"
#include "memory.h"

#ifdef HAVE_LIBCRYPTO
#include "db.h"

static unsigned int dh_gen_1024;
static unsigned char dh_prime_1024[];
static char *hex_to_string[256];

static BIGNUM *ircd_prime;
static BIGNUM *ircd_generator;

static int verify_is_hex(char *string)
{
    int l = strlen(string);
    char tmp[4] = { '\0', '\0', '\0', '\0' };
    int tmpidx = 0;

    if (l & 0x01) {		/* check if it's odd length */
	l++;
	tmp[tmpidx++] = '0';	/* pad with zero */
    }

    while (*string) {
	tmp[tmpidx++] = *string++;
	if (tmpidx == 2) {
    char *eptr;
    unsigned char x;

	    tmpidx = 0;

	    x = strtol(tmp, &eptr, 16);
	    if (*eptr != '\0') {
		return 0;
	    }
	}
    }
    return 1;
}

static void create_prime()
{
char buf[PRIME_BYTES_HEX];
int i;
int bufpos = 0;

    for (i = 0; i < PRIME_BYTES; i++) {
char *x = hex_to_string[dh_prime_1024[i]];
	while (*x)
	    buf[bufpos++] = *x++;
    }
    buf[bufpos] = '\0';

    ircd_prime = NULL;
    BN_hex2bn(&ircd_prime, buf);
    ircd_generator = BN_new();
    BN_set_word(ircd_generator, dh_gen_1024);
}

void dh_init(void)
{
    ERR_load_crypto_strings();

	ilog (L_INFO, "Importing prime...");
    create_prime();
	ilog (L_INFO, "Prime imported.");
}

int dh_generate_shared(struct dh_session *si, char *public_key)
{
    BIGNUM *tmp;
    int len;

    if (verify_is_hex(public_key) == 0 || !si || si->session_shared)
	return 0;

    tmp = NULL;
    BN_hex2bn(&tmp, public_key);
    if (!tmp)
	return 0;

    si->session_shared_length = DH_size(si->dh);
    si->session_shared = (char *) MyMalloc(DH_size(si->dh));
    len = DH_compute_key(si->session_shared, tmp, si->dh);
    BN_free(tmp);

    if (len < 0)
	return 0;

    si->session_shared_length = len;

    return 1;
}

struct dh_session *dh_start_session(void)
{
    struct dh_session *si = (struct dh_session *) MyMalloc(sizeof(struct dh_session));
    if (!si)
	abort();

    memset(si, 0, sizeof(struct dh_session));

    si->dh = DH_new();
    si->dh->p = BN_dup(ircd_prime);
    si->dh->g = BN_dup(ircd_generator);

    if (!DH_generate_key(si->dh)) {
	DH_free(si->dh);
	MyFree(si);
	return NULL;
    }

    return (void *) si;
}

void dh_end_session(struct dh_session *si)
{
    if (si->dh) {
	DH_free(si->dh);
	si->dh = NULL;
    }

    if (si->session_shared) {
	memset(si->session_shared, 0, si->session_shared_length);
	MyFree(si->session_shared);
	si->session_shared = NULL;
    }

    MyFree(si);
}

char *dh_get_s_public(char *buf, int maxlen, struct dh_session *si)
{
    char *tmp;

    if (!si || !si->dh || !si->dh->pub_key)
	return NULL;

    tmp = BN_bn2hex(si->dh->pub_key);
    if (!tmp)
	return NULL;

    if (strlen(tmp) + 1 > maxlen) {
	OPENSSL_free(tmp);
	return NULL;
    }
    strcpy(buf, tmp);
    OPENSSL_free(tmp);

    return buf;
}

int dh_get_s_shared(char *buf, int *maxlen, struct dh_session *si)
{
    if (!si || !si->session_shared || *maxlen < si->session_shared_length)
	return 0;

    *maxlen = si->session_shared_length;
    memcpy(buf, si->session_shared, *maxlen);

    return 1;
}

static char *hex_to_string[256] = {
    "00", "01", "02", "03", "04", "05", "06", "07",
    "08", "09", "0a", "0b", "0c", "0d", "0e", "0f",
    "10", "11", "12", "13", "14", "15", "16", "17",
    "18", "19", "1a", "1b", "1c", "1d", "1e", "1f",
    "20", "21", "22", "23", "24", "25", "26", "27",
    "28", "29", "2a", "2b", "2c", "2d", "2e", "2f",
    "30", "31", "32", "33", "34", "35", "36", "37",
    "38", "39", "3a", "3b", "3c", "3d", "3e", "3f",
    "40", "41", "42", "43", "44", "45", "46", "47",
    "48", "49", "4a", "4b", "4c", "4d", "4e", "4f",
    "50", "51", "52", "53", "54", "55", "56", "57",
    "58", "59", "5a", "5b", "5c", "5d", "5e", "5f",
    "60", "61", "62", "63", "64", "65", "66", "67",
    "68", "69", "6a", "6b", "6c", "6d", "6e", "6f",
    "70", "71", "72", "73", "74", "75", "76", "77",
    "78", "79", "7a", "7b", "7c", "7d", "7e", "7f",
    "80", "81", "82", "83", "84", "85", "86", "87",
    "88", "89", "8a", "8b", "8c", "8d", "8e", "8f",
    "90", "91", "92", "93", "94", "95", "96", "97",
    "98", "99", "9a", "9b", "9c", "9d", "9e", "9f",
    "a0", "a1", "a2", "a3", "a4", "a5", "a6", "a7",
    "a8", "a9", "aa", "ab", "ac", "ad", "ae", "af",
    "b0", "b1", "b2", "b3", "b4", "b5", "b6", "b7",
    "b8", "b9", "ba", "bb", "bc", "bd", "be", "bf",
    "c0", "c1", "c2", "c3", "c4", "c5", "c6", "c7",
    "c8", "c9", "ca", "cb", "cc", "cd", "ce", "cf",
    "d0", "d1", "d2", "d3", "d4", "d5", "d6", "d7",
    "d8", "d9", "da", "db", "dc", "dd", "de", "df",
    "e0", "e1", "e2", "e3", "e4", "e5", "e6", "e7",
    "e8", "e9", "ea", "eb", "ec", "ed", "ee", "ef",
    "f0", "f1", "f2", "f3", "f4", "f5", "f6", "f7",
    "f8", "f9", "fa", "fb", "fc", "fd", "fe", "ff"
};

static unsigned int dh_gen_1024 = 2;
static unsigned char dh_prime_1024[] = {
    0xF4, 0x88, 0xFD, 0x58, 0x4E, 0x49, 0xDB, 0xCD,
    0x20, 0xB4, 0x9D, 0xE4, 0x91, 0x07, 0x36, 0x6B,
    0x33, 0x6C, 0x38, 0x0D, 0x45, 0x1D, 0x0F, 0x7C,
    0x88, 0xB3, 0x1C, 0x7C, 0x5B, 0x2D, 0x8E, 0xF6,
    0xF3, 0xC9, 0x23, 0xC0, 0x43, 0xF0, 0xA5, 0x5B,
    0x18, 0x8D, 0x8E, 0xBB, 0x55, 0x8C, 0xB8, 0x5D,
    0x38, 0xD3, 0x34, 0xFD, 0x7C, 0x17, 0x57, 0x43,
    0xA3, 0x1D, 0x18, 0x6C, 0xDE, 0x33, 0x21, 0x2C,
    0xB5, 0x2A, 0xFF, 0x3C, 0xE1, 0xB1, 0x29, 0x40,
    0x18, 0x11, 0x8D, 0x7C, 0x84, 0xA7, 0x0A, 0x72,
    0xD6, 0x86, 0xC4, 0x03, 0x19, 0xC8, 0x07, 0x29,
    0x7A, 0xCA, 0x95, 0x0C, 0xD9, 0x96, 0x9F, 0xAB,
    0xD0, 0x0A, 0x50, 0x9B, 0x02, 0x46, 0xD3, 0x08,
    0x3D, 0x66, 0xA4, 0x5D, 0x41, 0x9F, 0x9C, 0x7C,
    0xBD, 0x89, 0x4B, 0x22, 0x19, 0x26, 0xBA, 0xAB,
    0xA2, 0x5E, 0xC3, 0x55, 0xE9, 0x2F, 0x78, 0xC7
};

#endif
